#include "UserWidgetAnimation.h"

#pragma region User Widget Animation

UUserWidgetAnimation::UUserWidgetAnimation() {
	bIsConstructedOnRunning = GIsRunning;
	PlayState = EPlayState::Stopped;
	Length = 0.0f;
	Speed = 1.0f;
	bIsLooping = false;
	CurrentTime = 0.0f;
	UserWidgetAnimationDataArray = TArray<FUserWidgetAnimationData>();
	UserWidgetAnimationOnStart = FUserWidgetAnimationOnStart();
	UserWidgetAnimationOnPause = FUserWidgetAnimationOnPause();
	UserWidgetAnimationOnStop = FUserWidgetAnimationOnStop();
	UserWidgetAnimationOnJumpTo = FUserWidgetAnimationOnJumpTo();
	UserWidgetAnimationOnEnd = FUserWidgetAnimationOnEnd();
}

void UUserWidgetAnimation::OnConstruct() {
	CalculateLengthFromUserWidgetAnimationDataArray();
}

void UUserWidgetAnimation::CalculateLengthFromUserWidgetAnimationDataArray()
{
	for (int32 i = 0; i < UserWidgetAnimationDataArray.Num(); i++) {
		if (UserWidgetAnimationDataArray[i].Length > Length) {
			Length = UserWidgetAnimationDataArray[i].Length;
		}
	}
}

UUserWidgetAnimation* UUserWidgetAnimation::CreateUserWidgetAnimation(
	const TArray<FUserWidgetAnimationData>& InUserWidgetAnimationDataArray)
{
	UUserWidgetAnimation* UserWidgetAnimation = NewObject<UUserWidgetAnimation>();
	UserWidgetAnimation->UserWidgetAnimationDataArray = InUserWidgetAnimationDataArray;
	UserWidgetAnimation->OnConstruct();
	return UserWidgetAnimation;
}

void UUserWidgetAnimation::UpdateOnTick(const float& DeltaSeconds)
{
	if (PlayState == EPlayState::PlayForward || PlayState == EPlayState::PlayReverse) {
		CalculateNewCurrentTime(DeltaSeconds);
		UpdateAnimation();
		CheckForAnimationEnd();
	}
}

void UUserWidgetAnimation::CalculateNewCurrentTime(const float& DeltaSeconds) {
	if (PlayState == EPlayState::PlayForward) {
		if (bIsLooping) {
			CurrentTime = FMath::Fmod(CurrentTime + (DeltaSeconds * Speed), Length);
		}
		else {
			CurrentTime = FMath::Clamp(CurrentTime + (DeltaSeconds * Speed), 0.0f, Length);
		}
	}
	else if(PlayState == EPlayState::PlayReverse) {
		if (bIsLooping) {
			CurrentTime = CurrentTime - (DeltaSeconds * Speed);
			int32 LoopCounter = 0;
			while (CurrentTime < 0.0f && LoopCounter < 10) {
				CurrentTime = Length - CurrentTime;
				LoopCounter++;
			}
		}
		else {
			CurrentTime = FMath::Clamp(CurrentTime - (DeltaSeconds * Speed), 0.0f, Length);
		}
	}
}

void UUserWidgetAnimation::UpdateAnimation() {
	for (int32 i = 0; i < UserWidgetAnimationDataArray.Num(); i++) {
		UserWidgetAnimationDataArray[i].UpdateAnimation(CurrentTime);
	}
}

void UUserWidgetAnimation::CheckForAnimationEnd() {
	if (!bIsLooping) {
		if ((PlayState == EPlayState::PlayForward && CurrentTime >= Length) ||
			(PlayState == EPlayState::PlayReverse && CurrentTime <= 0.0f)) 
		{
			StopAnimation();
		}
	}
}

void UUserWidgetAnimation::PlayAnimationForward(const float& InSpeed, const bool& bInIsLooping)
{
	PlayState = EPlayState::PlayForward;
	Speed = InSpeed;
	bIsLooping = bInIsLooping;
	UserWidgetAnimationOnStart.Broadcast(this,
		(PlayState == EPlayState::PlayForward ? EForward::Forward : EForward::Backward));
}

void UUserWidgetAnimation::PlayAnimationReverse(const float& InSpeed, const bool& bInIsLooping)
{
	if (PlayState == EPlayState::Stopped) {
		CurrentTime = Length;
	}
	PlayState = EPlayState::PlayReverse;
	Speed = InSpeed;
	bIsLooping = bInIsLooping;
	UserWidgetAnimationOnStart.Broadcast(this,
		(PlayState == EPlayState::PlayForward ? EForward::Forward : EForward::Backward));
}

void UUserWidgetAnimation::PauseAnimation()
{
	PlayState = EPlayState::Paused;
	UserWidgetAnimationOnPause.Broadcast(this);
}

void UUserWidgetAnimation::StopAnimation()
{
	if (PlayState == EPlayState::PlayForward || PlayState == EPlayState::PlayReverse) {
		UserWidgetAnimationOnEnd.Broadcast(this, 
			(PlayState == EPlayState::PlayForward ? EForward::Forward : EForward::Backward));
	}
	PlayState = EPlayState::Stopped;
	CurrentTime = 0.0f;
	UserWidgetAnimationOnStop.Broadcast(this);
}

void UUserWidgetAnimation::UpdateFromCurrentState()
{
	UpdateAnimation();
}

void UUserWidgetAnimation::JumpTo(const float& InTime)
{
	CurrentTime = InTime;
	UpdateAnimation(); 
	UserWidgetAnimationOnJumpTo.Broadcast(this, InTime);
}

void UUserWidgetAnimation::JumpToStart()
{
	JumpTo(0.0f);
}

void UUserWidgetAnimation::JumpToEnd()
{
	JumpTo(Length);
}

EPlayState UUserWidgetAnimation::GetPlayState() const
{
	return PlayState;
}

float UUserWidgetAnimation::GetLength() const
{
	return Length;
}

float UUserWidgetAnimation::GetSpeed() const
{
	return Speed;
}

void UUserWidgetAnimation::SetSpeed(const float& InSpeed)
{
	Speed = InSpeed;
}

bool UUserWidgetAnimation::GetIsLooping() const
{
	return bIsLooping;
}

void UUserWidgetAnimation::SetIsLooping(const bool& bInIsLooping)
{
	bIsLooping = bInIsLooping;
}

float UUserWidgetAnimation::GetCurrentTime() const
{
	return CurrentTime;
}

TArray<FUserWidgetAnimationData> UUserWidgetAnimation::GetUserWidgetAnimationDataArray() const
{
	return UserWidgetAnimationDataArray;
}

TArray<UWidget*> UUserWidgetAnimation::GetAnimationWidgets() const
{
	TArray<UWidget*> AnimationWidgets;
	for (int32 i = 0; i < UserWidgetAnimationDataArray.Num(); i++) {
		AnimationWidgets.Add(UserWidgetAnimationDataArray[i].Widget);
	}
	return AnimationWidgets;
}

void UUserWidgetAnimation::Tick(float DeltaTime)
{
	UpdateOnTick(DeltaTime);
}

bool UUserWidgetAnimation::IsTickable() const
{
	return bIsConstructedOnRunning;
}

TStatId UUserWidgetAnimation::GetStatId() const
{
	return UObject::GetStatID();
}

#pragma endregion