#pragma once

// Change YOURGAME_API to be the API of your game
// Change SetRenderTint() when not using the render tint files

#include "CoreMinimal.h"
#include "UObject/Object.h"
#include "Components/Widget.h"
#include "Tickable.h"
#include "UserWidgetAnimation.generated.h"

#pragma region Forward Declarations

class UUserWidgetAnimation;

#pragma endregion

#pragma region Enums

UENUM()
enum class EForward : uint8 {
	None,
	Forward,
	Backward
};

UENUM()
enum class EPlayState : uint8 {
	None,
	Stopped,
	Paused,
	PlayForward,
	PlayReverse
};

UENUM()
enum class ETransformType : uint8 {
	None,
	Opacity,
	TranslationX,
	TranslationY,
	ScaleX,
	ScaleY,
	Angle,
	ColorR,
	ColorB,
	ColorG,
	ColorA
};

#pragma endregion

#pragma region Delegates

DECLARE_DYNAMIC_MULTICAST_DELEGATE_TwoParams(FUserWidgetAnimationOnStart,
UUserWidgetAnimation*, UserWidgetAnimation, const EForward&, Forward);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_OneParam(FUserWidgetAnimationOnPause,
UUserWidgetAnimation*, UserWidgetAnimation);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_OneParam(FUserWidgetAnimationOnStop,
UUserWidgetAnimation*, UserWidgetAnimation);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_TwoParams(FUserWidgetAnimationOnJumpTo,
UUserWidgetAnimation*, UserWidgetAnimation, const float&, Time);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_TwoParams(FUserWidgetAnimationOnEnd,
UUserWidgetAnimation*, UserWidgetAnimation, const EForward&, Forward);

#pragma endregion

#pragma region User Widget Animation Key Frame Values

USTRUCT()
struct FUserWidgetAnimationKeyFrameValues {
	GENERATED_BODY()
public:
	UPROPERTY()
	TMap<int32, float> KeyFrameValues;
	UPROPERTY()
	float Length;
protected:
	UPROPERTY()
	TArray<int32> KeyFrameIntTimes;
	UPROPERTY()
	TArray<float> KeyFrameTimes;
	void CalculateKeyFrameIntTimes() {
		KeyFrameValues.GenerateKeyArray(KeyFrameIntTimes);
	}
	void SortKeyFrameValues() {
		KeyFrameIntTimes.Sort();
		TMap<int32, float> NewKeyFrameValues;
		for (int32 i = 0; i < KeyFrameIntTimes.Num(); i++) {
			NewKeyFrameValues.Add(KeyFrameIntTimes[i],
				KeyFrameValues.FindRef(KeyFrameIntTimes[i]));
		}
		KeyFrameValues = NewKeyFrameValues;
	}
	void CalculateKeyFrameTimes() {
		for (int32 i = 0; i < KeyFrameIntTimes.Num(); i++) {
			KeyFrameTimes.Add(((float)KeyFrameIntTimes[i]) / 1000.0f);
		}
	}
	void CalculateLengthFromKeyFrameTimes() {
		for (int32 i = 0; i < KeyFrameTimes.Num(); i++) {
			if (KeyFrameTimes[i]  > Length) {
				Length = KeyFrameTimes[i];
			}
		}
	}
public:
	FUserWidgetAnimationKeyFrameValues() {
		KeyFrameValues = TMap<int32, float>();
		Length = 0.0f;
	}
	FUserWidgetAnimationKeyFrameValues(const TMap<int32, float>& InKeyFrameValues) {
		KeyFrameValues = InKeyFrameValues;
		Length = 0.0f;
		CalculateKeyFrameIntTimes();
		SortKeyFrameValues();
		CalculateKeyFrameTimes();
		CalculateLengthFromKeyFrameTimes();
	}
	float GetKeyFrameValue(const float& Time) {
		if (KeyFrameTimes.Num() == 1) {
			return KeyFrameValues.FindRef(KeyFrameIntTimes[0]);
		}
		for (int32 i = 0; i < KeyFrameTimes.Num(); i++) {
			if ((Time <= KeyFrameTimes[i] && i == 0) ||
				(i == KeyFrameTimes.Num() - 1)) 
			{
				return KeyFrameValues.FindRef(KeyFrameIntTimes[i]);
			}
			else if(Time >= KeyFrameTimes[i] && Time < KeyFrameTimes[i+1])
			{
				return FMath::GetMappedRangeValueClamped(
					FVector2D(KeyFrameTimes[i], KeyFrameTimes[i + 1]),
					FVector2D(KeyFrameValues.FindRef(KeyFrameIntTimes[i]), KeyFrameValues.FindRef(KeyFrameIntTimes[i + 1])),
					Time);
			}
		}
		return 0.0f;
	}
};

#pragma endregion

#pragma region User Widget Animation Data

USTRUCT()
struct FUserWidgetAnimationData {
	GENERATED_BODY()
public:
	UPROPERTY()
	UWidget* Widget;
	UPROPERTY()
	TMap<ETransformType, FUserWidgetAnimationKeyFrameValues> UserWidgetAnimationKeyFrameValues;
	UPROPERTY()
	float Length;
protected:
	UPROPERTY()
	TArray<ETransformType> TransformTypes;
	void GenerateTransformTypes(){
		UserWidgetAnimationKeyFrameValues.GenerateKeyArray(TransformTypes);
	};
	void CalculateLength() {
		for (int32 i = 0; i < TransformTypes.Num(); i++) {
			if (UserWidgetAnimationKeyFrameValues.Find(TransformTypes[i])->Length > Length) {
				Length = UserWidgetAnimationKeyFrameValues.Find(TransformTypes[i])->Length;
			}
		}
	}
	void UpdateTransformFromTransformType(
		const ETransformType& TransformType, const float& Value)
	{
		if (Widget) {
			switch (TransformType) {
			case ETransformType::Opacity:
				Widget->SetRenderOpacity(Value);
				break;
			case ETransformType::TranslationX:
				Widget->SetRenderTranslation(FVector2D(Value, Widget->RenderTransform.Translation.Y));
				break;
			case ETransformType::TranslationY:
				Widget->SetRenderTranslation(FVector2D(Widget->RenderTransform.Translation.X, Value));
				break;
			case ETransformType::ScaleX:
				Widget->SetRenderScale(FVector2D(Value, Widget->RenderTransform.Scale.Y));
				break;
			case ETransformType::ScaleY:
				Widget->SetRenderScale(FVector2D(Widget->RenderTransform.Scale.X, Value));
				break;
			case ETransformType::Angle:
				Widget->SetRenderTransformAngle(Value);
				break;
			// Change the color transforms to use SetColorAndOpacity of UImage and UTextBlock
			// when not using the Widget Render Tint engine overrides.
			case ETransformType::ColorR:
				Widget->SetRenderTint(
					FLinearColor(
						Value,
						Widget->GetRenderTint().G,
						Widget->GetRenderTint().B,
						Widget->GetRenderTint().A
					)
				);
				break;
			case ETransformType::ColorG:
				Widget->SetRenderTint(
					FLinearColor(
						Widget->GetRenderTint().R,
						Value,
						Widget->GetRenderTint().B,
						Widget->GetRenderTint().A
					)
				);
				break;
			case ETransformType::ColorB:
				Widget->SetRenderTint(
					FLinearColor(
						Widget->GetRenderTint().R,
						Widget->GetRenderTint().G,
						Value,
						Widget->GetRenderTint().A
					)
				);
				break;
			case ETransformType::ColorA:
				Widget->SetRenderTint(
					FLinearColor(
						Widget->GetRenderTint().R,
						Widget->GetRenderTint().G,
						Widget->GetRenderTint().B,
						Value
					)
				);
				break;
			default:
				break;
			}
		}
	}
public:
	FUserWidgetAnimationData() {
		Widget = nullptr;
		UserWidgetAnimationKeyFrameValues = TMap<ETransformType, FUserWidgetAnimationKeyFrameValues>();
		Length = 0.0f;
	}
	FUserWidgetAnimationData(UWidget* InWidget,
		const TMap<ETransformType, FUserWidgetAnimationKeyFrameValues>& InUserWidgetAnimationKeyFrameValues)
	{
		Widget = InWidget;
		UserWidgetAnimationKeyFrameValues = InUserWidgetAnimationKeyFrameValues;
		Length = 0.0f;
		GenerateTransformTypes();
		CalculateLength();
	}
	void UpdateAnimation(const float& Time) {
		for (int32 i = 0; i < TransformTypes.Num(); i++) {
			UpdateTransformFromTransformType(
				TransformTypes[i],
				UserWidgetAnimationKeyFrameValues.Find(TransformTypes[i])->GetKeyFrameValue(Time));
		}
	}
};

#pragma endregion

#pragma region User Widget Animation

UCLASS()
class YOURGAME_API UUserWidgetAnimation : public UObject, public FTickableGameObject
{
	GENERATED_BODY()
protected:
	UPROPERTY()
	bool bIsConstructedOnRunning;
	UPROPERTY()
	EPlayState PlayState;
	UPROPERTY()
	float Length;
	UPROPERTY()
	float Speed;
	UPROPERTY()
	bool bIsLooping;
	UPROPERTY()
	float CurrentTime;
	UPROPERTY()
	TArray<FUserWidgetAnimationData> UserWidgetAnimationDataArray;
	void OnConstruct();
	void CalculateLengthFromUserWidgetAnimationDataArray();
	void UpdateOnTick(const float& DeltaSeconds);
	void CalculateNewCurrentTime(const float& DeltaSeconds);
	void UpdateAnimation();
	void CheckForAnimationEnd();
public:
	UPROPERTY()
	FUserWidgetAnimationOnStart UserWidgetAnimationOnStart;
	UPROPERTY()
	FUserWidgetAnimationOnPause UserWidgetAnimationOnPause;
	UPROPERTY()
	FUserWidgetAnimationOnStop UserWidgetAnimationOnStop;
	UPROPERTY()
	FUserWidgetAnimationOnJumpTo UserWidgetAnimationOnJumpTo;
	UPROPERTY()
	FUserWidgetAnimationOnEnd UserWidgetAnimationOnEnd;
	UUserWidgetAnimation();
	static UUserWidgetAnimation* CreateUserWidgetAnimation(
		const TArray<FUserWidgetAnimationData>& InUserWidgetAnimationDataArray);
	void PlayAnimationForward(const float& InSpeed = 1.0f, const bool& bInIsLooping = false);
	void PlayAnimationReverse(const float& InSpeed = 1.0f, const bool& bInIsLooping = false);
	void PauseAnimation();
	void StopAnimation();
	void UpdateFromCurrentState();
	void JumpTo(const float& InTime);
	void JumpToStart();
	void JumpToEnd();
	EPlayState GetPlayState() const;
	float GetLength() const;
	float GetSpeed() const;
	void SetSpeed(const float& InSpeed);
	bool GetIsLooping() const;
	void SetIsLooping(const bool& bInIsLooping);
	float GetCurrentTime() const;
	TArray<FUserWidgetAnimationData> GetUserWidgetAnimationDataArray() const;
	TArray<UWidget*> GetAnimationWidgets() const;
	virtual void Tick(float DeltaTime) override;
	virtual bool IsTickable() const override;
	virtual TStatId GetStatId() const override;
};

#pragma endregion